@@ -196,7 +196,14 @@ Kind getKind(const Decl *d, SymbolKind &kind) {
     // ccls extension
     kind = SymbolKind::Parameter;
     return Kind::Var;
-  case Decl::Var:
+  case Decl::Var: {
+    auto vd = cast<VarDecl>(d);
+    if (vd->isStaticDataMember()) {
+      kind = SymbolKind::Field;
+      return Kind::Var;
+    }
+    [[fallthrough]];
+  }
   case Decl::Decomposition:
     kind = SymbolKind::Variable;
     return Kind::Var;
@@ -358,7 +365,7 @@ const Decl *getAdjustedDecl(const Decl *d) {
   while (d) {
     if (auto *r = dyn_cast<CXXRecordDecl>(d)) {
       if (auto *s = dyn_cast<ClassTemplateSpecializationDecl>(r)) {
-        if (!s->getTypeAsWritten()) {
+        if (!s->isExplicitSpecialization()) {
           llvm::PointerUnion<ClassTemplateDecl *,
                              ClassTemplatePartialSpecializationDecl *>
               result = s->getSpecializedTemplateOrPartial();
@@ -534,7 +541,8 @@ public:
       name.replace(i, short_name.size(), qualified);
       def.short_name_offset = i + qualified.size() - short_name.size();
     }
-    def.short_name_size = short_name.size();
+    // name may be empty while short_name is not.
+    def.short_name_size = name.empty() ? 0 : short_name.size();
     for (int paren = 0; i; i--) {
       // Skip parentheses in "(anon struct)::name"
       if (name[i - 1] == ')')
@@ -663,7 +671,7 @@ public:
           rd->isInvalidDecl() || !validateRecord(rd))
         offset = -1;
       for (FieldDecl *fd : rd->fields()) {
-        int offset1 = offset < 0 ? -1 : offset + ctx->getFieldOffset(fd);
+        int offset1 = offset < 0 ? -1 : int(offset + ctx->getFieldOffset(fd));
         if (fd->getIdentifier())
           type.def.vars.emplace_back(getUsr(fd), offset1);
         else if (const auto *rt1 = fd->getType()->getAs<RecordType>()) {
@@ -946,11 +954,12 @@ public:
             }
       }
       [[fallthrough]];
+    case Decl::Enum:
     case Decl::Record:
-      if (auto *rd = dyn_cast<RecordDecl>(d)) {
+      if (auto *tag_d = dyn_cast<TagDecl>(d)) {
         if (type->def.detailed_name[0] == '\0' && info->short_name.empty()) {
           StringRef tag;
-          switch (rd->getTagKind()) {
+          switch (tag_d->getTagKind()) {
           case TTK_Struct:
             tag = "struct";
             break;
@@ -967,7 +976,7 @@ public:
             tag = "enum";
             break;
           }
-          if (TypedefNameDecl *td = rd->getTypedefNameForAnonDecl()) {
+          if (TypedefNameDecl *td = tag_d->getTypedefNameForAnonDecl()) {
             StringRef name = td->getName();
             std::string detailed = ("anon " + tag + " " + name).str();
             type->def.detailed_name = intern(detailed);
@@ -978,7 +987,7 @@ public:
             type->def.short_name_size = name.size();
           }
         }
-        if (is_def)
+        if (is_def && !isa<EnumDecl>(d))
           if (auto *ord = dyn_cast<RecordDecl>(origD))
             collectRecordMembers(*type, ord);
       }
@@ -1194,6 +1203,17 @@ public:
     return std::make_unique<MultiplexConsumer>(std::move(consumers));
   }
 };
+
+class IndexDiags : public DiagnosticConsumer {
+public:
+  llvm::SmallString<64> message;
+  void HandleDiagnostic(DiagnosticsEngine::Level level,
+    const clang::Diagnostic &info) override {
+    DiagnosticConsumer::HandleDiagnostic(level, info);
+    if (message.empty())
+      info.FormatDiagnostic(message);
+  }
+};
 } // namespace
 
 const int IndexFile::kMajorVersion = 21;
@@ -1243,7 +1263,7 @@ void init() {
                                        g_config->index.multiVersionBlacklist);
 }
 
-std::vector<std::unique_ptr<IndexFile>>
+IndexResult
 index(SemaManager *manager, WorkingFiles *wfiles, VFS *vfs,
       const std::string &opt_wdir, const std::string &main,
       const std::vector<const char *> &args,
@@ -1272,7 +1292,7 @@ index(SemaManager *manager, WorkingFiles *wfiles, VFS *vfs,
       ci->getPreprocessorOpts().addRemappedFile(filename, bufs.back().get());
     }
 
-  DiagnosticConsumer dc;
+  IndexDiags dc;
   auto clang = std::make_unique<CompilerInstance>(pch);
   clang->setInvocation(std::move(ci));
   clang->createDiagnostics(&dc, false);
@@ -1346,7 +1366,10 @@ index(SemaManager *manager, WorkingFiles *wfiles, VFS *vfs,
     return {};
   }
 
-  std::vector<std::unique_ptr<IndexFile>> result;
+  IndexResult result;
+  result.n_errs = (int)dc.getNumErrors();
+  // clang 7 does not implement operator std::string.
+  result.first_error = std::string(dc.message.data(), dc.message.size());
   for (auto &it : param.uid2file) {
     if (!it.second.db)
       continue;
@@ -1383,7 +1406,7 @@ index(SemaManager *manager, WorkingFiles *wfiles, VFS *vfs,
         entry->dependencies[llvm::CachedHashStringRef(intern(path))] =
             file.mtime;
     }
-    result.push_back(std::move(entry));
+    result.indexes.push_back(std::move(entry));
   }
 
   return result;
