@@ -11,6 +11,7 @@
 
 #include <clang/Sema/CodeCompleteConsumer.h>
 #include <clang/Sema/Sema.h>
+#include <llvm/ADT/Twine.h>
 
 #if LLVM_VERSION_MAJOR < 8
 #include <regex>
@@ -370,7 +371,7 @@ void buildItem(const CodeCompletionResult &r, const CodeCompletionString &ccs,
           continue;
         }
         out[i].textEdit.newText +=
-            "${" + std::to_string(out[i].parameters_.size()) + ":" + text + "}";
+            ("${" + Twine(out[i].parameters_.size()) + ":" + text + "}").str();
         out[i].insertTextFormat = InsertTextFormat::Snippet;
       } else if (kind != CodeCompletionString::CK_Informative) {
         out[i].textEdit.newText += text;
@@ -450,9 +451,23 @@ public:
       buildItem(r, *ccs, ls_items);
 
       for (size_t j = first_idx; j < ls_items.size(); j++) {
-        if (g_config->client.snippetSupport &&
-            ls_items[j].insertTextFormat == InsertTextFormat::Snippet)
-          ls_items[j].textEdit.newText += "$0";
+        std::string &s = ls_items[j].textEdit.newText;
+        if (!g_config->client.snippetSupport) {
+          if (s.size()) {
+            // Delete non-identifier parts.
+            if (s.back() == '(' || s.back() == '<')
+              s.pop_back();
+            else if (s.size() >= 2 && !s.compare(s.size() - 2, 2, "()"))
+              s.resize(s.size() - 2);
+          }
+        } else if (ls_items[j].insertTextFormat == InsertTextFormat::Snippet) {
+          if (!g_config->completion.placeholder) {
+            // foo(${1:int a}, ${2:int b}) -> foo($1)$0
+            auto p = s.find("${"), q = s.rfind('}');
+            s.replace(p, q - p + 1, "$1");
+          }
+          s += "$0";
+        }
         ls_items[j].priority_ = ccs->getPriority();
         if (!g_config->completion.detailedLabel) {
           ls_items[j].detail = ls_items[j].label;
@@ -565,13 +580,14 @@ void MessageHandler::textDocument_completion(CompletionParam &param,
         if (!consumer->from_cache) {
           cache.withLock([&]() {
             cache.path = path;
+            cache.line = buffer_line;
             cache.position = begin_pos;
             cache.result = consumer->ls_items;
           });
         }
       };
 
-  if (cache.isCacheValid(path, begin_pos)) {
+  if (cache.isCacheValid(path, buffer_line, begin_pos)) {
     CompletionConsumer consumer(ccOpts, true);
     cache.withLock([&]() { consumer.ls_items = cache.result; });
     callback(&consumer);
