@@ -22,6 +22,7 @@
 #include <llvm/Support/Threading.h>
 
 #include <chrono>
+#include <inttypes.h>
 #include <mutex>
 #include <shared_mutex>
 #include <thread>
@@ -38,6 +39,12 @@ struct PublishDiagnosticParam {
   std::vector<Diagnostic> diagnostics;
 };
 REFLECT_STRUCT(PublishDiagnosticParam, uri, diagnostics);
+
+constexpr char index_progress_token[] = "index";
+struct WorkDoneProgressCreateParam {
+  const char *token = index_progress_token;
+};
+REFLECT_STRUCT(WorkDoneProgressCreateParam, token);
 } // namespace
 
 void VFS::clear() {
@@ -67,7 +74,8 @@ void standaloneInitialize(MessageHandler &, const std::string &root);
 namespace pipeline {
 
 std::atomic<bool> g_quit;
-std::atomic<int64_t> loaded_ts{0}, pending_index_requests{0}, request_id{0};
+std::atomic<int64_t> loaded_ts{0}, request_id{0};
+IndexStats stats;
 int64_t tick = 0;
 
 namespace {
@@ -195,9 +203,6 @@ bool indexer_Parse(SemaManager *completion, WorkingFiles *wfiles,
     return false;
   auto &request = *opt_request;
   bool loud = request.mode != IndexMode::OnChange;
-  struct RAII {
-    ~RAII() { pending_index_requests--; }
-  } raii;
 
   // Dummy one to trigger refresh semantic highlight.
   if (request.path.empty()) {
@@ -207,6 +212,9 @@ bool indexer_Parse(SemaManager *completion, WorkingFiles *wfiles,
     return false;
   }
 
+  struct RAII {
+    ~RAII() { stats.completed++; }
+  } raii;
   if (!matcher.matches(request.path)) {
     LOG_IF_S(INFO, loud) << "skip " << request.path;
     return false;
@@ -330,17 +338,9 @@ bool indexer_Parse(SemaManager *completion, WorkingFiles *wfiles,
       return true;
     } while (0);
 
-  if (loud) {
-    std::string line;
-    if (LOG_V_ENABLED(1)) {
-      line = "\n ";
-      for (auto &arg : entry.args)
-        (line += ' ') += arg;
-    }
-    LOG_S(INFO) << (deleted ? "delete " : "parse ") << path_to_index << line;
-  }
-
   std::vector<std::unique_ptr<IndexFile>> indexes;
+  int n_errs = 0;
+  std::string first_error;
   if (deleted) {
     indexes.push_back(std::make_unique<IndexFile>(request.path, "", false));
     if (request.path != path_to_index)
@@ -353,8 +353,12 @@ bool indexer_Parse(SemaManager *completion, WorkingFiles *wfiles,
         remapped.emplace_back(path_to_index, content);
     }
     bool ok;
-    indexes = idx::index(completion, wfiles, vfs, entry.directory,
-                         path_to_index, entry.args, remapped, no_linkage, ok);
+    auto result =
+        idx::index(completion, wfiles, vfs, entry.directory, path_to_index,
+                   entry.args, remapped, no_linkage, ok);
+    indexes = std::move(result.indexes);
+    n_errs = result.n_errs;
+    first_error = std::move(result.first_error);
 
     if (!ok) {
       if (request.id.valid()) {
@@ -367,6 +371,21 @@ bool indexer_Parse(SemaManager *completion, WorkingFiles *wfiles,
     }
   }
 
+  if (loud || n_errs) {
+    std::string line;
+    SmallString<64> tmp;
+    SmallString<256> msg;
+    (Twine(deleted ? "delete " : "parse ") + path_to_index).toVector(msg);
+    if (n_errs)
+      msg += (" error:" + Twine(n_errs) + " " + first_error).toStringRef(tmp);
+    if (LOG_V_ENABLED(1)) {
+      msg += "\n ";
+      for (const char *arg : entry.args)
+        (msg += ' ') += arg;
+    }
+    LOG_S(INFO) << std::string_view(msg.data(), msg.size());
+  }
+
   for (std::unique_ptr<IndexFile> &curr : indexes) {
     std::string path = curr->path;
     if (!matcher.matches(path)) {
@@ -605,13 +624,13 @@ void mainLoop() {
 
   SemaManager manager(
       &project, &wfiles,
-      [&](std::string path, std::vector<Diagnostic> diagnostics) {
+      [](const std::string &path, std::vector<Diagnostic> diagnostics) {
         PublishDiagnosticParam params;
         params.uri = DocumentUri::fromPath(path);
-        params.diagnostics = diagnostics;
+        params.diagnostics = std::move(diagnostics);
         notify("textDocument/publishDiagnostics", params);
       },
-      [](RequestId id) {
+      [](const RequestId &id) {
         if (id.valid()) {
           ResponseError err;
           err.code = ErrorCode::InternalError;
@@ -632,7 +651,9 @@ void mainLoop() {
   handler.manager = &manager;
   handler.include_complete = &include_complete;
 
+  bool work_done_created = false, in_progress = false;
   bool has_indexed = false;
+  int64_t last_completed = 0;
   std::deque<InMessage> backlog;
   StringMap<std::deque<InMessage *>> path2backlog;
   while (true) {
@@ -682,6 +703,44 @@ void mainLoop() {
       }
     }
 
+    int64_t completed = stats.completed.load(std::memory_order_relaxed);
+    if (completed != last_completed) {
+      if (!work_done_created) {
+        WorkDoneProgressCreateParam param;
+        request("window/workDoneProgress/create", param);
+        work_done_created = true;
+      }
+
+      int64_t enqueued = stats.enqueued.load(std::memory_order_relaxed);
+      if (completed != enqueued) {
+        if (!in_progress) {
+          WorkDoneProgressParam param;
+          param.token = index_progress_token;
+          param.value.kind = "begin";
+          param.value.title = "indexing";
+          notify("$/progress", param);
+          in_progress = true;
+        }
+        int64_t last_idle = stats.last_idle.load(std::memory_order_relaxed);
+        WorkDoneProgressParam param;
+        param.token = index_progress_token;
+        param.value.kind = "report";
+        param.value.message =
+            (Twine(completed - last_idle) + "/" + Twine(enqueued - last_idle))
+                .str();
+        param.value.percentage =
+            100.0 * (completed - last_idle) / (enqueued - last_idle);
+        notify("$/progress", param);
+      } else if (in_progress) {
+        stats.last_idle.store(enqueued, std::memory_order_relaxed);
+        WorkDoneProgressParam param;
+        param.token = index_progress_token;
+        param.value.kind = "end";
+        notify("$/progress", param);
+        in_progress = false;
+      }
+    }
+
     if (did_work) {
       has_indexed |= indexed;
       if (g_quit.load(std::memory_order_relaxed))
@@ -706,8 +765,9 @@ void standalone(const std::string &root) {
   WorkingFiles wfiles;
   VFS vfs;
   SemaManager manager(
-      nullptr, nullptr, [&](std::string, std::vector<Diagnostic>) {},
-      [](RequestId id) {});
+      nullptr, nullptr,
+      [](const std::string &, const std::vector<Diagnostic> &) {},
+      [](const RequestId &id) {});
   IncludeComplete complete(&project);
 
   MessageHandler handler;
@@ -724,16 +784,16 @@ void standalone(const std::string &root) {
     int entries = 0;
     for (auto &[_, folder] : project.root2folder)
       entries += folder.entries.size();
-    printf("entries: %5d\n", entries);
+    printf("entries:   %4d\n", entries);
   }
   while (1) {
     (void)on_indexed->dequeueAll();
-    int pending = pending_index_requests;
+    int64_t enqueued = stats.enqueued, completed = stats.completed;
     if (tty) {
-      printf("\rpending: %5d", pending);
+      printf("\rcompleted: %4" PRId64 "/%" PRId64, completed, enqueued);
       fflush(stdout);
     }
-    if (!pending)
+    if (completed == enqueued)
       break;
     std::this_thread::sleep_for(std::chrono::milliseconds(100));
   }
@@ -744,8 +804,9 @@ void standalone(const std::string &root) {
 
 void index(const std::string &path, const std::vector<const char *> &args,
            IndexMode mode, bool must_exist, RequestId id) {
-  pending_index_requests++;
-  index_request->pushBack({path, args, mode, must_exist, id},
+  if (!path.empty())
+    stats.enqueued++;
+  index_request->pushBack({path, args, mode, must_exist, std::move(id)},
                           mode != IndexMode::Background);
 }
 
@@ -789,7 +850,7 @@ void notifyOrRequest(const char *method, bool request,
   for_stdout->pushBack(output.GetString());
 }
 
-static void reply(RequestId id, const char *key,
+static void reply(const RequestId &id, const char *key,
                   const std::function<void(JsonWriter &)> &fn) {
   rapidjson::StringBuffer output;
   rapidjson::Writer<rapidjson::StringBuffer> w(output);
@@ -802,7 +863,7 @@ static void reply(RequestId id, const char *key,
     w.Null();
     break;
   case RequestId::kInt:
-    w.Int(atoll(id.value.c_str()));
+    w.Int64(atoll(id.value.c_str()));
     break;
   case RequestId::kString:
     w.String(id.value.c_str(), id.value.size());
@@ -817,11 +878,12 @@ static void reply(RequestId id, const char *key,
   for_stdout->pushBack(output.GetString());
 }
 
-void reply(RequestId id, const std::function<void(JsonWriter &)> &fn) {
+void reply(const RequestId &id, const std::function<void(JsonWriter &)> &fn) {
   reply(id, "result", fn);
 }
 
-void replyError(RequestId id, const std::function<void(JsonWriter &)> &fn) {
+void replyError(const RequestId &id,
+                const std::function<void(JsonWriter &)> &fn) {
   reply(id, "error", fn);
 }
 } // namespace pipeline
