@@ -90,6 +90,10 @@ struct ProjectProcessor {
         i++;
       return true;
     }
+    if (arg == "-Xclang") {
+      i++;
+      return true;
+    }
     return exclude_args.count(arg) ||
            any_of(exclude_globs,
                   [&](const GlobPattern &glob) { return glob.match(arg); });
@@ -233,7 +237,7 @@ bool appendToCDB(const std::vector<const char *> &args) {
   return args.size() && StringRef("%compile_commands.json") == args[0];
 }
 
-std::vector<const char *> getFallback(const std::string path) {
+std::vector<const char *> getFallback(const std::string &path) {
   std::vector<const char *> argv{"clang"};
   if (sys::path::extension(path) == ".h")
     argv.push_back("-xobjective-c++-header");
@@ -429,7 +433,9 @@ void Project::loadDirectory(const std::string &root, Project::Folder &folder) {
       // If workspace folder is real/ but entries use symlink/, convert to
       // real/.
       entry.directory = realPath(cmd.Directory);
+      entry.directory.push_back('/');
       normalizeFolder(entry.directory);
+      entry.directory.pop_back();
       doPathMapping(entry.directory);
       entry.filename =
           realPath(resolveIfRelative(entry.directory, cmd.Filename));
@@ -516,9 +522,10 @@ Project::Entry Project::findEntry(const std::string &path, bool can_redirect,
         auto it = folder.path2entry_index.find(path);
         if (it != folder.path2entry_index.end()) {
           Project::Entry &entry = folder.entries[it->second];
-          exact_match = entry.filename == path;
-          if ((match = exact_match || can_redirect) || entry.compdb_size) {
-            // best->compdb_size is >0 for a compdb entry, 0 for a .ccls entry.
+          // best->compdb_size is >0 for an explicit compile_commands.json
+          // entry, 0 for an implicit entry.
+          exact_match = entry.compdb_size > 0 && entry.filename == path;
+          if ((match = exact_match || can_redirect) && entry.compdb_size) {
             best_compdb_folder = &folder;
             best = &entry;
           }
@@ -547,11 +554,16 @@ Project::Entry Project::findEntry(const std::string &path, bool can_redirect,
     if (!best) {
       // Infer args from a similar path.
       int best_score = INT_MIN;
+      auto [lang, header] = lookupExtension(path);
       for (auto &[root, folder] : root2folder)
         if (StringRef(path).startswith(root))
           for (const Entry &e : folder.entries)
             if (e.compdb_size) {
               int score = computeGuessScore(path, e.filename);
+              // Decrease score if .c is matched against .hh
+              auto [lang1, header1] = lookupExtension(e.filename);
+              if (lang != lang1 && !(lang == LanguageId::C && header))
+                score -= 30;
               if (score > best_score) {
                 best_score = score;
                 best_compdb_folder = &folder;
@@ -591,7 +603,7 @@ Project::Entry Project::findEntry(const std::string &path, bool can_redirect,
   return ret;
 }
 
-void Project::index(WorkingFiles *wfiles, RequestId id) {
+void Project::index(WorkingFiles *wfiles, const RequestId &id) {
   auto &gi = g_config->index;
   GroupMatch match(gi.whitelist, gi.blacklist),
       match_i(gi.initialWhitelist, gi.initialBlacklist);
